;;; db-convert.el
;;; Michael D. Ernst <mernst@theory.lcs.mit.edu>
;;; This file is part of EDB, the Emacs database.

;;; Primitive routines for changing the field structure of a database.

;;; That is, this takes a database and rearranges the field order, converts
;;; fields from one type to another, adds or removes fields, etc.
;;; There are two parts to the following code:  the user interface and the
;;; actual conversion.



;;; Variables used dynamically; avoid compiler messages about free variables.
;; Should this be?  Perhaps get rid of it.
(defvar new-db-fieldnames)
;; Should this be?  Perhaps get rid of it.
(defvar computed-functions)


(require 'database)
(provide 'db-convert)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Notes
;;;

;; For the database being created, there is an alist of two-element lists
;; of (fieldname creation-method).  The fieldname is only used in
;; eventually setting the fieldlist.

;; Should the lists contain an atom or a function (in which case there will
;; need to be more editing commands when editing this graphically)?  If I
;; use funcall, the first argument can be a function or an atom.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; User interface
;;;

;; The screen is split into two columns (not necessarily two windows), with
;; the old database structure on the left and the new database structure on
;; the right.  From each record in the original database, one record is
;; created in the new database; by default, the new record is identical to
;; the old.

;; In the new database structure, field F is specified as one of:
;; * literal.  Field F of every record is set to that value.
;; * field reference.  Field F is set to the value of the specified field
;;   in the corresponding old record.  User specifies names, but internally
;;   numbers are manipulated.  (Conversion functions should take names or
;;   numbers to simplify calling from a Emacs Lisp.)
;; * result of function call.  Specified are the function, what arguments
;;   should be passed to it, and which of the results is desired.  The
;;   arguments are fields of the original record (or the entire record).
;;   Specifying one of the results permits a function which does a lot of
;;   work (and computes the values of a lot of fields in the new record) to
;;   be called only once per record; its results are memoized and reused
;;   later as needed.  These numbers are displayed in base 1 but probably
;;   manipulated internally in base 0.

;; This will be just as much work as the sort interface was -- probably more.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Conversion
;;;

;; Currently this doesn't create a new database; it destructively changes
;; the format of the old one.  I suspect this is the desired behavior.

(defun db-convert (database new-db-fieldnames)
  ;; New-db-fieldnames isn't actually a straight fieldlist.
  (let ((new-db-fieldnames (db-compile-down-creation-methods new-db-fieldnames
							     database)))
    (maplinks (function db-convert-link)
	      database))

  ;; Change stuff like fieldlist, accessor functions, etc.
 
  ;; changed from database-set-fieldnames per Henry Thompson
  (db-set-fieldname-vars database
	(mapcar (function car) new-db-fieldnames))

  )

;; This has to be a function of one argument so that it can be called in maplinks.
;; Perhaps I should be using maplinks-macro instead, then.  Probably.

;; Probably instead of setting the link record, create a new database.
(defun db-convert-link (link)
  (link-set-record link (db-convert-record (link-record link))))

;; Uses dynamically-bound new-db-fieldnames.
;; Dynamically binds computed-functions.
(defun db-convert-record (record)
  ;; Return a new record.
  (let ((new-fieldlist new-db-fieldnames)
	(new-record (make-vector (length new-db-fieldnames) nil))
	(this-field-no 0)
	computed-functions)
    (while new-fieldlist
      (aset new-record this-field-no
	    (db-compute-field-value (second (car new-fieldlist)) record))

      ;; ...

      (setq new-fieldlist (cdr new-fieldlist))
      (setq this-field-no (1+ this-field-no)))
    new-record
    ))

;; Uses the "computed-functions" dynamic variable.
;; Perhaps get rid of the "result" variable.
(defun db-compute-field-value (creation-method old-record)
  (cond	((stringp creation-method)
	 creation-method)
	((numberp creation-method)
	 (aref old-record creation-method))
	((consp creation-method)
	 (let ((function (car creation-method))
	       (result-no (cdr creation-method))
	       result-found)
	   (if result-no
	       (setq result-found (assoc function computed-functions)))
	   (if result-found
	       (nth result-no result-found)
	     (let ((result (apply (car function)
				  (mapcar (function (lambda (slot-no)
						      (aref old-record slot-no)))
					  (cdr function)))))
	       (if result-no
		   (progn
		     (setq computed-functions
			   (cons (cons function result) computed-functions))
		     (nth (1- result-no) result))
		 result)))))
	(t (error "Unknown creation method %s" creation-method))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Processing user-set creation-method lists
;;;

;; Input is a list of creation methods, except that:
;;  * nil, or no specification of creation method at all (ie, list contains
;; an atom instead of a list) can be used for the empty string
;;  * in function specifications, may use just the function specification
;; if no record-no; may use function alone (ie, no parens) if applied to no args



;; Creation methods:
;;  * string:  use that string as the value.
;;  * number:  like atom, but for efficiency
;;  * cons cell:  car is a function specification, cdr a result-number.
;;	Function specification is a list of the function and the argument
;;	numbers; call function on those elts of the old record, where -1
;;	means the entire record.  If result-number is nil, use entire
;;	result.  Otherwise use (nth result-number result); this permits one
;;	function to set several values.  These values are memoized.


(defun db-compile-down-creation-methods (creation-methods database)
  (mapcar (function (lambda (method)
		      (db-compile-down-creation-method method database)))
	  creation-methods))


(defun db-compile-down-creation-method (creation-method database)
  (cond
   ((atom creation-method)
    (list creation-method ""))
   ((and (listp creation-method) (= 1 (length creation-method)))
    (list (car creation-method) ""))
   (t
    (let ((method (cadr creation-method)))
      (list (car creation-method)
	    (cond
	     ((null method)
	      "")
	     ((or (stringp method) (numberp method))
	      method)
	     ((and (atom method)
		   (fieldname->fieldnumber method database)))
	     (t
	      (db-compile-down-creation-method-function method))
	     )				; end of cond
	    )))
   )					; end of cond
  )

(defun db-compile-down-creation-method-function (method)
  (let (function args result-no)
    (if (atomp method)
	(setq function method)
      (progn
	(if (or (numberp (cdr method)) (consp (car method)))
	    (setq function (car method)
		  result-no (cdr method))
	  (setq function method))
	(if (listp function)
	    (setq args (cdr function)
		  function (car function)))))
    ;; check that the function is a function

    (if (not (fboundp function))
	(error "%s has no function definition." function))
    ;; was name-or-number->fieldnumber
    (setq args (mapcar (function fieldname->fieldnumber)
		       args))
    (cons (cons function args) result-no)))


