;;; db-types.el
;;; Michael D. Ernst <mernst@theory.lcs.mit.edu>
;;; This file is part of EDB, the Emacs database.


;;; Library of types for EDB database fields.

;;; This file contains predefined types.  For efficiency, they're defined
;;; in terms of the displayspec abstraction instead of via format strings.
;;; Improvements and additions are welcome.  Predefined types for dates and
;;; times can be found in db-time.el.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Contents
;;;

;;; Contents
;;; To Do
;;; Variables
;;; Define Displaytype
;;; Define Enumeration Displaytypes
;;; Numbers
;;; Booleans
;;; Strings
;;; Last Names
;;; First Names
;;; Full Names
;;; Places


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; To Do
;;;


;; Perhaps change many of these (those I know will never be multiline) to
;; non-rectangularp.


;; The question is whether we want this to be separate or the information
;; to be contained in the displayspec.
;; In displayspec:
;;  * no need for extra indirection to get the information
;;  * easier to customize (eg to provide more specific help information).
;; In datatype:
;;  * less repetition of information, slightly more space-efficient.


;; In any event I want to be able to specify the type either in
;; db-fieldnames or in field specifications.

;; I should permit field specifications to not even include type information.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;

(defvar db-use-completing-read t
  "*Non-nil if EDB should use completing-read when dealing with enumerated types.
Otherwise, on erroneous enumeration input, EDB shows the possible completions
but doesn't correct the input.")

(defvar db-enum-ignore-case t
  "If non-nil, assume that any association lists created from
enumeration types input names are to include both upper and lower case
versions of the name, if distinct.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Define displaytype
;;;

;; WARNING!  We may want the type field of the displayspec to be different
;; from the argument here; this implies we want to be able to specify a
;; type either elsewhere or in optstring.  Easier for me if it's not in
;; optstring, of course.  Now, why make the user deal with optstring at
;; all?  Why not let him use an alist or something if he wants?  Answer:
;; because I'm too lazy just now for that.  [Uh, who uses the type field?
;; Wouldn't we want the type field to reflect the first argument?  Do we
;; just want to permit inheritance here?]

;; Perhaps permit inheritance based on inheritedtype; perhaps permit
;; setting of type field dependent on actualtype

;; Perhaps permit typename to be different from an abbreviation permitted
;; in field specifications.

(defun define-displaytype-from-optstring (typename optstring)
  "Define a displaytype named TYPENAME according to OPTSTRING.
TYPENAME is a symbol or string and OPTSTRING is the optional parameters
part of a display specification string."
  (if (stringp typename)
      (setq typename (intern typename)))
  (let ((displayspec (make-displayspec-from-type-and-options nil optstring)))
    (define-displaytype-from-displayspec typename displayspec)))

;; Typename must be a symbol.  This is an internal function.
(defun define-displaytype-from-displayspec (typename displayspec)
  "Define a displaytype named TYPENAME (a symbol) with the default DISPLAYSPEC.
DISPLAYSPEC may also be a typename symbol itself."
  (setq db-displaytypes (cons (cons typename displayspec) db-displaytypes)))
(put 'define-displaytype-from-displayspec 'lisp-indent-hook 1)

(defun define-recordfieldtype-from-recordfieldspec (typename recordfieldspec)
  "Define a recordfieldtype named TYPENAME (a symbol) with the default RECORDFIELDSPEC.
DISPLAYSPEC may also be a typename symbol itself.
After this call, `recordfieldtype->recordfieldspec' called with argument
TYPENAME returns the proper record field specification."
  (setq db-recordfieldtypes (cons (cons typename recordfieldspec) db-recordfieldtypes)))
(put 'define-recordfieldtype-from-recordfieldspec 'lisp-indent-hook 1)

;; Provide functions to easily create aliases to existing types

(fset 'define-displaytype-alias 'define-displaytype-from-displayspec)
(make-obsolete 'define-displaytype-alias 'define-displaytype-from-displayspec)
(fset 'define-recordfieldtype-alias 'define-recordfieldtype-from-recordfieldspec)
(make-obsolete 'define-recordfieldtype-alias 'define-recordfieldtype-from-recordfieldspec)

(defun define-type-alias (alias typename)
  "Make symbol ALIAS refer to the same displaytype and recordfieldtype as TYPENAME."
  (define-displaytype-from-displayspec alias typename)
  (define-recordfieldtype-from-recordfieldspec alias typename))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Define enumeration displaytypes
;;;

;; An enumeration displaytype is used for fields whose values are one of a
;; fixed set of alternatives.  There are two types of alternatives:  those
;; consisting of a single character (say, M or F for gender) and those
;; consisting of a whole string.  The latter, multicharacter alternative
;; type is nicknamed an "enum" type.

;; special syntax for defining alternatives
(defun define-alternative-one-char-displaytype (typename alternatives optstring)
  (error "define-alternative-one-char-displaytype isn't defined yet")
  )

;; When creating an enum, the alternatives are kept in a database local
;; variable, dbc-enums-alist, under an alist keyed by the type name.

(defun define-alternative-multi-char-displaytype (typename alternatives &optional optstring)
  "Make NAME (a symbol or string) an enumerated type according to ALTERNATIVES.
ALTERNATIVES is a list of strings or cons cells, the car of which is an
acceptable alternative for input, and the cdr of which is the corresponding
display representation.
Optional argument OPTSTRING is a displayspec option string.

There should only be one display representation for each alternative, and
the display representation must also be a valid input string for the 
alternative."
  (if (stringp typename)
      (setq typename (intern typename)))
  (let ((ds (if optstring
		(make-displayspec-from-type-and-options
		 nil optstring)
	      (make-displayspec))))
    (db-enum-set-alist typename (db-enum-make-alist alternatives))
    (displayspec-set-indent ds nil)
    (displayspec-set-display->actual ds (make-string->enum typename))
    (displayspec-set-actual->display ds (make-enum->string typename))
    (define-displaytype-from-displayspec typename ds))
  (let ((rs (make-recordfieldspec)))
    (recordfieldspec-set-type rs typename)
    (recordfieldspec-set-default-value rs "")
    (recordfieldspec-set-order-fn rs (make-enum-member-orderer typename))
    (recordfieldspec-set-sort-fn rs (make-enum-member-sorter typename))
    ;; With this value, we should have
    ;; (displayspec-set-match-actual->display ds 'string-match-actual->display)
    ;; (displayspec-set-match-display->actual ds 'string-match-display->actual)
    ;; in the displayspec, I believe.
    (recordfieldspec-set-match-function rs (function string-match-function))
    (recordfieldspec-set-help-info rs (make-enumerated-help typename))
    (define-recordfieldtype-from-recordfieldspec typename rs)))

;; A much shorter (nicer) function name: define-enum-type
(fset 'define-enum-type 'define-alternative-multi-char-displaytype)


;;; alternatives abstraction
;;; [AKS - 3/4/93]

(defun db-enum-name-in (alt)
  "The input name for the current enum alternative."
  (if (consp alt)
      (car alt)
    alt))
(defun db-enum-name-out (alt)
  "The string used for output of the current enum alternative."
  (if (consp alt)
      (cdr alt)
    alt))
(proclaim-inline db-enum-name-in db-enum-name-out)


(defun db-enum-make-alist (enum-list)
  "Return an association list for the ENUM-LIST of enumeration type
alternatives.  Each element of ENUM-LIST is a string or a cons of
\(INPUT-NAME . OUTPUT-NAME\)."

  (let (alt-alist)
    (mapcar 
     (function (lambda (alt)
	(let ((iname (db-enum-name-in alt))
	      (oname (db-enum-name-out alt)))
	  (if (not (assoc iname alt-alist))
	      (setq alt-alist (cons (cons iname oname) alt-alist))))))
     enum-list)				;end mapcar
    ;; The order is significant in make-enumerated-help.
    ;; [It's not clear whether that's a feature or a bug.]
    (nreverse alt-alist)))


(defun db-enum-set-alist (type alist &optional no-error)
  "Given an enum TYPE and its alternatives ALIST, save them as part of
the local database variables.  You can do this only once, unless optional
third arg is non-nil.

ALIST is not a simple alternatives list, but an association list of the
sort returned by `db-enum-make-alist'."

  (let ((enum-alist (database-get-local 'dbc-enums-alist dbc-database t)))
    (if (and (assoc type enum-alist)
	     (not no-error))
	(error "Enum type \"%s\" is multiply defined!" (symbol-name type))
      (database-set-local 'dbc-enums-alist dbc-database
			  (cons (cons type alist) enum-alist)
			  t))))


(defun db-enum-get-alist (type &optional no-error)
  "Given an enum TYPE name, return its alist of alternatives."
  (let* ((enums-alist (database-get-local 'dbc-enums-alist dbc-database t))
	 (this-alist (assoc type enums-alist)))
    (cond (this-alist
	   (cdr this-alist))
	  ((not no-error)
	   (error "Enum type %s not defined." (symbol-name type))))))
	   
;;; make-enumerated-help  [Added by AKS, 10/13/92].

(defun make-enumerated-help (type)
  "For the enum TYPE, return a form which, when evaled, displays
all its alternatives."

  (let* ((buf (get-buffer-create " *temp*"))
	 (standard-output buf)
	 (enum-alist (db-enum-get-alist type)))
    (princ (format "%s: An enumerated type.  " (symbol-name type)))
    (display-completion-list (all-completions "" enum-alist))
    (save-excursion
      (set-buffer buf)
      (prog1 (buffer-string)
	(kill-buffer buf)))))


(defun make-enum->string (type)
  "Create a function which will convert an enum value of type TYPE into its
printed representation.  Allows mixed case matching."
  (` (lambda (enum-val) (db-enum->string (quote (, type)) enum-val))))


(defun db-enum->string (type value)
  "Given an enum TYPE, convert the VALUE to its printed name."
  (let ((alist (db-enum-get-alist type)))
    (if value
	(or (cdr-safe (assoc value alist))
	    (and db-enum-ignore-case
		 (or (cdr-safe (assoc (downcase value) alist))
		     (cdr-safe (assoc (upcase value) alist))
		     (cdr-safe (assoc (capitalize value) alist))
		     ""))
	    "")
      "")
    ))


(defun make-string->enum (type)
  "Create a form which evals to a function which will convert its string
argument into one of the alternatives of TYPE; if ambiguities exist, they
are resolved interactively."
  (` (lambda (enum-item)
       (if enum-item
	   (db-enum-do-completions (quote (, type)) enum-item)
	 ""))))


(defun db-enum-do-completions (type item)
  "Given an enum TYPE and a string ITEM, return ITEM if it is a
valid selection, otherwise see if it completes to one of the valid
selections, and show the possible completions."
  
  (let* ((alt-alist (db-enum-get-alist type)))
    (if (assoc item alt-alist)
	item
      (let* ((completion-ignore-case db-enum-ignore-case)
	     (try-comp (try-completion item alt-alist)))
	(if (and try-comp
		 (assoc try-comp alt-alist))
	    try-comp
	  (progn
	    (dbf-set-this-field-text (or try-comp ""))
	    (if db-use-completing-read
		(completing-read (format "Enter \"%s\" enum value (? for list): " type)
				 alt-alist nil t (or try-comp ""))
	      (progn
		;; no completing read, so display the options
		(with-electric-help-maybe
		 (display-completion-list
		  (all-completions (or try-comp "") alt-alist)))
		;;       (if db-use-ehelp-p
		;; 	  (with-electric-help
		;; 	   (function (lambda ()
		;; 	      (display-completion-list
		;; 	       (all-completions (or try-comp "") alt-alist)))))
		;; 	(with-output-to-temp-buffer "*Help*"
		;; 	  (display-completion-list
		;; 	   (all-completions (or try-comp "") alt-alist))))
		(error (if try-comp
			   "Not unique."
			 "No match."))))
	    )				;end progn
	  ))				;end if, let
      ))				;end if, let
  )					;end db-enum-do-completions


(defun make-enum-member-orderer (type)
  "Return an ordering function for the enum type TYPE."
  (make-enum-member-orderer-internal type -1 0 1))
(proclaim-inline make-enum-member-orderer)

(defun make-enum-member-sorter (type)
  "Return a sorting function for the enum type TYPE."
  (make-enum-member-orderer-internal type t nil nil))
(proclaim-inline make-enum-member-sorter)

(defun make-enum-member-orderer-internal (type less-value equal-value greater-value)
  "Given an enum type TYPE, create an ordering function which compares two
items in the enum type's alternatives list.  The resulting function returns
LESS-VALUE if its first argument precedes its second argument in the
alternative list, EQUAL-VALUE if they're equal or neither appears, and
GREATER-VALUE if the first follows the second.

If one of the arguments doesn't appear in the alternatives list, the other
is considered to precede it."
  (` (lambda (item1 item2)
       (if (equal item1 item2)
	   (, equal-value)
	 (let ((alt-alist (db-enum-get-alist (, type)))
	       (result (, equal-value)))
	   (while alt-alist
	     (cond ((equal (db-enum-name-out (car alt-alist)) item1)
		    (setq result (, less-value)
			  alt-alist nil))
		   ((equal (db-enum-name-out (car alt-alist)) item2)
		    (setq result (, greater-value)
			  alt-alist nil))
		   (t
		    (setq alt-alist (cdr alt-alist)))))
	   result)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Numbers
;;;

;; I may wish to permit a numeric type that's stored as a string, not a
;; number; not because that will make conversion to display and storage
;; types easier (actually, moot) but because it will allow searching the
;; number like a string.


;; Ought I provide special handling of unspecified numbers?  Perhaps later;
;; that will require dealing with computation, too.

(let ((ds (make-displayspec)))
  (displayspec-set-indent ds nil)
  (displayspec-set-actual->display ds (function integer->string))
  (displayspec-set-display->actual ds (function string->integer))
  (define-displaytype-from-displayspec 'integer ds))
(let ((rs (make-recordfieldspec)))
  (recordfieldspec-set-type rs 'integer)
  (recordfieldspec-set-default-value rs 0)
  (recordfieldspec-set-actual->stored rs (function integer->string))
  (recordfieldspec-set-stored->actual rs (function string->integer))
  (recordfieldspec-set-order-fn rs (function integer-order))
  (recordfieldspec-set-sort-fn rs (function <))
  (recordfieldspec-set-match-function rs (function =))
  (recordfieldspec-set-help-info rs "An integer.")
  (define-recordfieldtype-from-recordfieldspec 'integer rs))

(let ((ds (make-displayspec)))
  (displayspec-set-indent ds nil)
  (displayspec-set-actual->display ds (function integer-or-nil->string))
  (displayspec-set-display->actual ds (function string->integer-or-nil))
  (define-displaytype-from-displayspec 'integer-or-nil ds))
(let ((rs (make-recordfieldspec)))
  (recordfieldspec-set-type rs 'integer)
  (recordfieldspec-set-default-value rs 0)
  (recordfieldspec-set-actual->stored rs (function integer-or-nil->string))
  (recordfieldspec-set-stored->actual rs (function string->integer-or-nil))
  (recordfieldspec-set-order-fn rs (function integer-or-nil-order-nil-greatest))
  ;; (recordfieldspec-set-sort-fn rs (function <))
  (recordfieldspec-set-match-function rs (function equal))
  (recordfieldspec-set-help-info rs "An integer, or nil.")
  (define-recordfieldtype-from-recordfieldspec 'integer-or-nil rs))

;; The following are defined in util-mde.el:
;;  string->integer-or-nil
;;  integer-or-nil->string
;;  string-or-nil->integer-or-nil

;; ;; Deal with requirement to take extra arguments.
;; (defun db-integer->string (i &rest ignore)
;;   (integer->string i))
;; (defun db-string->integer (s &rest ignore)
;;   (string->integer s))
;; (defun db-integer-or-nil->string (i &rest ignore)
;;   (integer-or-nil->string i))
;; (defun db-string->integer-or-nil (s &rest ignore)
;;   (string->integer-or-nil s))
;; (proclaim-inline db-integer->string db-string->integer
;; 		 db-integer-or-nil->string db-string->integer-or-nil)

(defun integer-order (a b)
  "Return -1, 0, or 1 depending on whether A < B, A = B, or A > B."
  (cond ((= a b) 0)
	((< a b) -1)
	(t 1)))
(proclaim-inline integer-order)

(defun integer-or-nil-order-nil-greatest (a b)
  "Like integer-order, but nil is treated as greater than any integer.
This puts nil after integers in an increasing list."
  (cond ((and a b)
	 (integer-order a b))
	(a -1)
	(b 1)
	(t 0)))

(defun integer-or-nil-order-nil-least (a b)
  "Like integer-order, but nil is treated as smaller than any integer.
This puts nil after integers in a decreasing list."
  (cond ((and a b)
	 (integer-order a b))
	(a 1)
	(b -1)
	(t 0)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Booleans
;;;

;; There could be a one-char boolean as well; implement it later.

(let ((ds (make-displayspec)))
  (displayspec-set-indent ds nil)
  (displayspec-set-min-width ds 3)
  (displayspec-set-max-width ds 3)
  (displayspec-set-actual->display ds (function boolean->yes-no-string))
  (displayspec-set-display->actual ds (function yes-no-string->boolean))
  (define-displaytype-from-displayspec 'yes-no ds))
(let ((rs (make-recordfieldspec)))
  (recordfieldspec-set-type rs 'boolean)
  (recordfieldspec-set-default-value rs nil)
  (recordfieldspec-set-actual->stored rs (function boolean->yn-string))
  (recordfieldspec-set-stored->actual rs (function yn-string->boolean))
  (recordfieldspec-set-order-fn rs (function boolean-order-function))
  (recordfieldspec-set-sort-fn rs (function boolean-lessp))
  (recordfieldspec-set-match-function rs (function eq))
  (recordfieldspec-set-help-info rs "A boolean value.")
  (define-recordfieldtype-from-recordfieldspec 'boolean rs))

(defun boolean->yes-no-string (boolean)
  (if boolean "Yes" "No"))
(proclaim-inline boolean->yes-no-string)

(defun yes-no-string->boolean (yes-or-no)
  (let ((downcased-yes-or-no (downcase yes-or-no)))
    (cond ((string= "yes" downcased-yes-or-no)
	   t)
	  ((string= "no" downcased-yes-or-no)
	   nil)
	  (t
	   (error "`%s' is not `Yes' or `No'." yes-or-no)))))

;; (defun db-boolean->yes-no-string (boolean &rest ignore)
;;   (boolean->yes-no-string boolean))
;; (defun db-yes-no-string->boolean (yes-or-no)
;;   (yes-no-string->boolean yes-or-no))
;; (proclaim-inline db-boolean->yes-no-string db-yes-no-string->boolean)

(defun boolean->yn-string (boolean)
  (if boolean "Y" "N"))
(proclaim-inline boolean->yn-string)

(defun yn-string->boolean (y-or-n)
  (string= "y" (downcase y-or-n)))
(proclaim-inline yn-string->boolean)

(defun string-is-yn-p (string)
  (and (= 1 (length string))
       (or (string= "Y" string)
	   (string= "y" string)
	   (string= "N" string)
	   (string= "n" string))))

;; t < nil so that in the "increasing" ordering "true" things occur before
;; false ones.  (This is somewhat arbitrary.)
(defun boolean-order-function (bool1 bool2)
  (cond ((or (and bool1 bool2)
	     (not (or bool1 bool2)))
	 0)
	(bool1
	 -1)
	(t
	 1)))

(defun boolean-lessp (bool1 bool2)
  (and (not bool1) bool2))
(proclaim-inline boolean-lessp)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Strings and variants
;;;

;; Type: string

(let ((ds (make-displayspec)))
  (displayspec-set-max-height ds nil)
  (displayspec-set-indent ds t)
  ;; use quote instead of function because of actual->display-call
  (displayspec-set-match-actual->display ds 'string-match-actual->display)
  (displayspec-set-match-display->actual ds 'string-match-display->actual)
  (define-displaytype-from-displayspec 'string ds))
(let ((rs (make-recordfieldspec)))
  (recordfieldspec-set-type rs 'string)
  (recordfieldspec-set-default-value rs "")
  (recordfieldspec-set-order-fn rs (function string-order-ci))
  (recordfieldspec-set-sort-fn rs (function string-lessp-ci))
  (recordfieldspec-set-match-function rs (function string-match-function))
  ;; String is the default type; I'm not sure I want
  ;; help-info for it.
  ;; (recordfieldspec-set-help-info rs "A string.")
  (define-recordfieldtype-from-recordfieldspec 'string rs))

;; Type: one-line-string

(let ((ds (copy-displayspec (displaytype->displayspec 'string))))
  (displayspec-set-max-height ds 1)
  (displayspec-set-indent ds nil)
  (define-displaytype-from-displayspec 'one-line-string ds))
;; This recordfieldspec is gratuitous.  Isn't it?
(let ((rs (copy-recordfieldspec (recordfieldtype->recordfieldspec 'string))))
  (recordfieldspec-set-type rs 'one-line-string)
  (define-recordfieldtype-from-recordfieldspec 'one-line-string rs))

;; Type: string-or-nil

(let ((ds (copy-displayspec (displaytype->displayspec 'string))))
  (displayspec-set-actual->display ds (function string-or-nil->string))
  (define-displaytype-from-displayspec 'string-or-nil ds))
(let ((rs (copy-recordfieldspec (recordfieldtype->recordfieldspec 'string))))
  (recordfieldspec-set-type rs 'string-or-nil)
  (recordfieldspec-set-order-fn rs (function string-or-nil-order-ci))
  (recordfieldspec-set-sort-fn rs (function string-or-nil-lessp-ci))
  (recordfieldspec-set-match-function rs (function string-or-nil-match-function))
  (define-recordfieldtype-from-recordfieldspec 'string-or-nil rs))

;; Type: nil-or-string

(let ((ds (copy-displayspec (displaytype->displayspec 'string-or-nil))))
  (displayspec-set-display->actual ds (function string->nil-or-string))
  (define-displaytype-from-displayspec 'nil-or-string ds))
(let ((rs (copy-recordfieldspec (recordfieldtype->recordfieldspec 'string-or-nil))))
  (recordfieldspec-set-type rs 'nil-or-string)
  (define-recordfieldtype-from-recordfieldspec 'nil-or-string rs))

;; Type: one-line-string-or-nil

(let ((ds (copy-displayspec (displaytype->displayspec 'one-line-string))))
  (displayspec-set-actual->display ds (function string-or-nil->string))
  (define-displaytype-from-displayspec 'one-line-string-or-nil ds))
(let ((rs (copy-recordfieldspec (recordfieldtype->recordfieldspec 'one-line-string))))
  (recordfieldspec-set-type rs 'one-line-string-or-nil)
  (recordfieldspec-set-order-fn rs (function string-or-nil-order-ci))
  (recordfieldspec-set-sort-fn rs (function string-or-nil-lessp-ci))
  (recordfieldspec-set-match-function rs (function string-or-nil-match-function))
  (define-recordfieldtype-from-recordfieldspec 'one-line-string-or-nil rs))

;; Helping functions for type string

(defun string-lessp-ci (string1 string2)
  "Case-insensitive version of string-lessp."
  (string-lessp (downcase string1) (downcase string2)))
(proclaim-inline string-lessp-ci)

(defun string-order-ci (string1 string2)
  "Return -1, 0, or 1 depending on whether STRING1 is lexicographically less
than, equal to, or greater than STRING2.  Case-insensitive."
  (let ((s1 (downcase string1))
	(s2 (downcase string2)))
    (cond ((string= s1 s2) 0)
	  ((string-lessp s1 s2) -1)
	  (t 1))))

;; Pattern is a list of 'regexp and a regexp or a list of 'string, a
;; regexp, and a string.
(defun make-regexp-pattern (regexp)
  (list 'regexp regexp))
(fset 'regexp-pattern-regexp 'second)
(defun make-string-pattern (string &optional regexp)
  (list 'string (or regexp (regexp-quote string)) string))
(fset 'string-pattern-regexp 'second)
(fset 'string-pattern-string 'third)

(defun string-match-function (pattern string)
  ;; The second element of pattern is a regexp whether the pattern is a
  ;; string or a regexp.
  ;; However, I also deal with the case where pattern is incorrectly a string.
  (string-match (if (stringp pattern) pattern (second pattern))
		string))
(proclaim-inline string-match-function)

;; Return a pattern
(defun string-match-display->actual (string)
  (if (string-match dbm-string-regexp-prefix string)
      (make-regexp-pattern (substring string (match-end 0)))
    (make-string-pattern string)))

(defun string-match-actual->display (pattern)
  (cond ((eq (car pattern) 'string)
	 (string-pattern-string pattern))
	((eq (car pattern) 'regexp)
	 (concat dbm-string-regexp-prefix-string
		 (regexp-pattern-regexp pattern)))
	(t
	 (error "string-match-actual->display:  bad pattern %s" pattern))))

;; Helping functions for type string-or-nil

(defun string-or-nil->string (s-o-n)
  (or s-o-n ""))
(proclaim-inline string-or-nil->string)

(defun string-or-nil-lessp-ci (s-o-n-1 s-o-n-2)
  (string-lessp-ci (string-or-nil->string s-o-n-1)
		   (string-or-nil->string s-o-n-2)))
(proclaim-inline string-or-nil-lessp-ci)

(defun string-or-nil-order-ci (s-o-n-1 s-o-n-2)
  (string-order-ci (or s-o-n-1 "") (or s-o-n-2 "")))
(proclaim-inline string-or-nil-order-ci)

(defun string-or-nil-match-function (pattern s-o-n)
  (string-match (if (stringp pattern) pattern (if pattern (cadr pattern) ""))
		(or s-o-n "")))
(proclaim-inline string-or-nil-match-function)

;; Helping functions for type nil-or-string

(defun string->nil-or-string (string)
  (if (empty-string-p string)
      nil
    string))
(proclaim-inline string->nil-or-string) 


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Last names
;;;

;; Note stringency of equality test:  capitalization and spacing matter.
(defun order-last-names (last-name-1 last-name-2)
  "Return -1, 0, or 1 depending on whether LAST-NAME-1 is lexographically
less than, equal to, or greater than LAST-NAME-2."
  (cond ((string-equal last-name-1 last-name-2)
	 0)
	((string-lessp (make-name-comparable last-name-1)
		       (make-name-comparable last-name-2))
	 -1)
	(t
	 1)))

;; The telephone book orders Mc between Mb and Md, but ignores spaces and
;; capitalization.
(defun make-name-comparable (last-name)
  (downcase (string-remove-regexp "[ ']" last-name)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; First names
;;;

;; change this setq back to a defvar later
(defvar nicknames
  '(
    ("Abraham" "Abe")
    ("Charles" "Charlie")
    ("David" "Dave")
    ("Gerald" "Gerry")
    ("James" "Jamey" "Jamie")
    ("Leonard" "Lenny")
    ("Michael" "Mike")
    ("Richard" "Dick" "Dicky" "Rick" "Ricky" "Rico")
    ("Robert" "Bob")
    ("Theodore" "Ted")
    )
  "List of nicknames which are not prefixes of the full name.
I add to this only when I need another one, so it's by no means comprehensive.")

(defun same-first-name-p (fname1 fname2)
  (or (string-equal fname1 fname2)
      (if (< (length fname1) (length fname2))
	  (nicknamep fname1 fname2)
	(nicknamep fname2 fname1))))
 
(defun nicknamep (nickname fullname)
  (or (string-match (regexp-quote nickname) fullname)
      (string-member nickname
		     ;; I think this string-split is utterly unnecessary.
		     ;; [That was for a particular application, I believe.]
		     (cdr (assoc (car (string-split-first-word fullname))
				 nicknames)))))

(defun order-first-names (first-name-1 first-name-2)
  (let ((first-middle-1 (string-split-first-word first-name-1))
	(first-middle-2 (string-split-first-word first-name-2)))
    (cond ((or
	    ;; This first test isn't necessary.
	    ;; [Is it?]
	    (same-first-name-p (car first-middle-1) (car first-middle-2))
	    (same-first-name-p first-name-1 first-name-2))
	   0)
	  ((string-lessp (car first-middle-1) (car first-middle-2))
	   -1)
	  (t
	   1))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Full names
;;;

(defun standardize-name (name)
  ;; remove leading honorifics
  (if (string-match (concat "^\\(\\(Dr\\|Mrs?\\|Prof\\|1?Lt\\)\\.?"
			    "\\|Miss\\|Ensign\\|Doctor\\) ") name)
      (setq name (substring name (match-end 0))))
  ;; remove trailing gubbish
  (setq name (substring name
			0 (string-match ",? \\(PhD\\|MD\\|Esq\\.?\\|Esquire\\)$" name)))
  ;; add periods after initials
  (while (string-match "\\(^\\| \\)[A-Z]\\([ ,]\\|$\\)" name)
    (setq name (concat (substring name 0 (match-beginning 2))
		       "."
		       (substring name (match-beginning 2)))))
  name)

;; preferred suffixes
(defvar jr-assoc-list
  '(("3d" . "3rd")))

;; return list of (suffixless-name suffix)
(defun name->name-jr (name)
  (if (string-match ",? +\\(jr\\.?\\|sr\\.?\\|3r?d\\|[45]th\\|I+\\)$" name)
      (list (substring name 0 (match-beginning 0))
	    (let ((jr (substring name (match-beginning 1))))
	      (or (cdr (assoc (downcase jr) jr-assoc-list))
		  jr)))
    (list name "")))

;; return list of (fname lname jr)
(defun name->first-last-jr (fullname)
  (let* ((name-jr (name->name-jr (standardize-name fullname)))
	 (split-names (string-split-last-word (car name-jr)
					      "\\(La\\|de\\) [A-Z-]+")))
    (append split-names (cdr name-jr))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Places
;;;

(defvar UK-postal-code-regexp
  "[A-Z][0-9][A-Z] ?[0-9][A-Z][0-9]")

(defvar zip-code-regexp
  "[0-9][0-9][0-9][0-9][0-9]\\(-?[0-9][0-9][0-9][0-9]\\)?")

(defvar postal-code-regexp
  (mapconcat
   (function identity)
   (list
    ;; US zip code
    zip-code-regexp
    ;; UK postal code
    UK-postal-code-regexp
    ;; other
    "[0-9][0-9A-Z]*")
   "\\|"))
			
(defun postal-code-p (string)
  (string-match postal-code-regexp string))
(proclaim-inline postal-code-p)

;; Full state-alist and associated functions added by
;; Alan Stebbens, UCSB, Sep 18 '92.

(defvar state-alist
  '(
    ( "AK" . "Alaska" )
    ( "AL" . "Alabama" )
    ( "AR" . "Arkansas" )
    ( "AZ" . "Arizona" )
    ( "CA" . "California" )
    ( "CO" . "Colorado" )
    ( "CT" . "Conneticut" )
    ( "DC" . "District of Columbia" )
    ( "DE" . "Deleware" )
    ( "FL" . "Florida" )
    ( "GA" . "Georgia" )
    ( "GU" . "Guam" )
    ( "HI" . "Hawaii" )
    ( "ID" . "Idaho" )
    ( "IL" . "Illinois" )
    ( "IN" . "Indiana" )
    ( "IA" . "Iowa" )
    ( "KS" . "Kansas" )
    ( "KY" . "Kentucky" )
    ( "LA" . "Louisiana" )
    ( "MA" . "Massachusetts" )
    ( "MD" . "Maryland" )
    ( "ME" . "Maine" )
    ( "MI" . "Michigan" )
    ( "MN" . "Minnessota" )
    ( "MS" . "Mississippi" )
    ( "MO" . "Missouri" )
    ( "MT" . "Montana" )
    ( "NE" . "Nebraska" )
    ( "NH" . "New Hampshire" )
    ( "NC" . "North Carolina" )
    ( "ND" . "North Dakota" )
    ( "NJ" . "New Jersey" )
    ( "NM" . "New Mexico" )
    ( "NV" . "Nevada" )
    ( "NY" . "New York" )
    ( "OH" . "Ohio" )
    ( "OK" . "Oklahoma" )
    ( "OR" . "Oregon" )
    ( "PA" . "Pennsylvania" )
    ( "PR" . "Puerto Rico" )
    ( "RI" . "Rhode Island" )
    ( "SC" . "South Carolina" )
    ( "SD" . "South Dakota" )
    ( "TX" . "Texas" )
    ( "TN" . "Tennessee" )
    ( "UT" . "Utah" )
    ( "VA" . "Virginia" )
    ( "VT" . "Vermont" )
    ( "WA" . "Washington" )
    ( "WI" . "Wisconson" )
    ( "WV" . "West Virginia" )
    ( "WY" . "Wyoming" )
    )
  "An alist of ABBREV and FULLNAME of each of the United States, and its
territories.  Used by \"full-state-name\" and \"abbreviate-state\"."
  )

(defun statep (string)
  "Return non-nil if STRING is a valid state abbreviation."
  (full-state-name string))

(defun full-state-name (abbrev)
  "Return the full state name for ABBREV, or nil if not found."
  (and (= 2 (length abbrev))
       ;; (let ((elt (assoc (upcase abbrev) state-alist)))
       ;;   (if elt (cdr elt)))
       (cdr (assoc (upcase abbrev) state-alist))))

(defun abbreviate-state (state)
  "Return the postal abbreviation for STATE, or nil if not found."
  (car (rassoc (capitalize state) state-alist)))

;; (defun abbreviate-state (state)
;;   "Return the postal abbreviation for STATE."
;;   (let ((alist state-alist)
;; 	(key (capitalize state)))
;;     (while (and alist
;; 		(not (string-equal key (cdar alist))))
;;       (setq alist (cdr alist)))
;;     (if alist (caar alist))))


(defvar country-list
  '("Canada"
    "Israel"
    "Switzerland"
    "Japan"
    "West Germany"
    "United States"
    "Zambia"))

(defun countryp (string)
  (string-member (capitalize string) country-list))
(proclaim-inline countryp)


;; Dates and times moved to db-time.el
