;;; Emacs Lisp utilities
;;; Michael Ernst <mernst@lcs.mit.edu>

(require 'util-clmde)
(provide 'util-mde)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Bug fixes
;;;

;; Prevent users of the standard byte-compiler, or users of the new one who
;; haven't yet loaded bytecomp-runtime.el, from barfing on these macros.
;; These could probably all be in one test, but this is more satisfying to
;; paranoid people.

;; This didn't work on all systems when the defuns and defmacros weren't at
;; top level.
(defmacro proclaim-inline-hack (&rest args)
  "No-op for non-users of the Zawinski byte-compiler."
  )
(if (not (fboundp 'proclaim-inline))
    (fset 'proclaim-inline 'proclaim-inline-hack))
(defmacro proclaim-notinline-hack (&rest args)
  "No-op for non-users of the Zawinski byte-compiler."
  )
(if (not (fboundp 'proclaim-notinline))
    (fset 'proclaim-notinline 'proclaim-notinline-hack))
(defun make-obsolete-hack (fn new)
  "No-op for non-users of the Zawinski byte-compiler."
  )
(if (not (fboundp 'make-obsolete))
    (fset 'make-obsolete 'make-obsolete-hack))
(if (not (fboundp 'inline))
    (progn
      (fset 'inline 'progn)
      (put 'inline 'lisp-indent-hook 0)))
(defun byte-compile-warn-hack (&rest args)
  "No-op for non-users of the Zawinski byte-compiler."
  )
(if (not (fboundp 'byte-compile-warn))
    (fset 'byte-compile-warn 'byte-compile-warn-hack))


;; In the standard 18.55 through 18.58 distributions, backquote has a bug.
;; Unfortunately, this fix may trounce on other, correct, implementations
;; of backquote, breaking them.  I don't know how to correct this, but you
;; can prevent the fix from being loaded by setting dont-fix-backquote.
;; I personally use a backquote implementation that doesn't happen to
;; define the function bq-splicequote.
(defvar dont-fix-backquote nil)
;; These defvars are here only to quiet the byte-compiler.
(defvar tailmaker)
(defvar state)
(if (not dont-fix-backquote)
    (progn
      (require 'backquote)
      (defun bq-splicequote (form)
	(setq tailmaker (list form (list 'quote tailmaker)))
	(setq state 'append))))


;; String-equal permits symbols, whose print names are used instead.  One
;; would hope that it would err when passed nil, or at least that it
;; wouldn't return t for (string-equal () "nil").  Therefore, it should be
;; avoided.
(make-obsolete 'string-equal 'equal)

;; Emacs 19 compatibility.
(if (not (fboundp 'buffer-disable-undo))
    (fset 'buffer-disable-undo 'buffer-flush-undo))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Numbers
;;;

(defmacro %-base-1 (x y)
  "Like %, but if (% x y) = 0, (%-base-1 x y) = y."
  (` (let* ((y-value (, y))
	    (result (% (, x) y-value)))
       (if (zerop result) y-value result))))
(fset 'mod-base-1 '%-base-1)

(defun power (x y)
  "Simplistic x^y for integral x, y > 0."
  (let ((result 1))
    (while (> y 0)
      (setq result (* x result)
	    y (1- y)))
    result))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Characters
;;;

(fset 'charp (function integerp))
;; (defun charp (char)
;;   "Return t if CHAR is a character, nil otherwise."
;;   (integerp char))

;; (defmacro charp (object)
;;   "Return t if CHAR is a character, nil otherwise."
;;   (` (integerp (, object))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Strings
;;;

;; (defun empty-string-p (string)
;;   "Return t if its argument (a string) has length 0, nil otherwise.
;; Don't apply this to nil, as the result will be nil."
;;   (string-equal "" string))
(defmacro empty-string-p (string)
  "Return t if STRING has length 0, nil otherwise.
Don't apply this to nil, as the result will be nil."
  (` (string-equal "" (, string))))

(defun empty-string-or-nil-p (string-or-nil)
  "Return t if its argument is nil or a zero-length string, nil otherwise."
  (or (not string-or-nil)
      (string-equal "" string-or-nil)))

(defmacro string-shorter-p (string1 string2)
  "Return t if STRING1 is shorter than STRING2, nil otherwise."
  (` (< (length (, string1)) (length (, string2)))))

;; This needs to remain a function so it can be passed as an argument.
(defun string-shorter-p-function (string1 string2)
  "Return t if STRING1 is shorter than STRING2, nil otherwise.
Good for passing to `sort' in order to sort strings by length."
  (< (length string1) (length string2)))

(defmacro string-longer-p (string1 string2)
  "Return t if STRING1 is longer than STRING2, nil otherwise."
  (` (> (length (, string1)) (length (, string2)))))

;; This needs to remain a function so it can be passed as an argument.
(defun string-longer-p-function (string1 string2)
  "Return t if STRING1 is longer than STRING2, nil otherwise.
Good for passing to sort in order to sort strings by length."
  (> (length string1) (length string2)))

(defun right-justify (string fieldwidth)
  (let ((slen-gap (- fieldwidth (length string))))
    (if (plusp slen-gap)
	(concat (make-string slen-gap ? ) string)
      string)))

;; It would be nice to make the delimiter an optional argument to these

;; A syntax table implementation would be too complicated, so hardcode space
;; and tab.
(defun string-split-last-word (splitee &optional exceptions delimiter)
  "Return list of two strings (all-but-last-word last-word).
If there is only one word, return (SPLITEE \"\").
The result strings can be concatenated to return the original string,
with the exception of some number (at least one) of spaces and tabs,
and possibly a comma immediately preceding them.
Optional arg EXCEPTIONS, if non-nil, is a regexp (containing spaces or tabs)
which, if found at the end of SPLITEE, should be considered a single word.
Optional arg DELIMITER, if non-nil, is used instead of the default word
delimiter.  It should be a regexp."
  (if (not delimiter) (setq delimiter ",?[ \t]+"))
  (if (or (and exceptions
	       (string-match (concat delimiter "\\(" exceptions "\\)$") splitee))
	  (string-match (concat delimiter "\\([a-zA-Z0-9'-]+\\)$" splitee)))
      (list (substring splitee 0 (match-beginning 0))
	    (substring splitee (match-beginning 1)))
    (list splitee "")))

;; maybe what I really want is string-to-word-list
(defun string-split-first-word (splitee &optional delimiter)
  "Return list of strings (first-word remaining-words).
Argument SPLITEE is split at the first sequence of spaces and tabs.
Optional arg DELIMITER, if non-nil, is used instead of the default word
delimiter.  It should be a regexp."
  (if (string-match (or delimiter "[ \t]+") splitee)
      (list (substring splitee 0 (match-beginning 0))
	    (substring splitee (match-end 0)))
    (list splitee "")))

;; Is it necessary to return the argument at the end?  I think that the
;; side-effects will propagate back in any case.
(defun string-substitute-substring-same-length (new old-regexp string)
  (let ((i 0)
	j
	max-j)
    (while (setq i (string-match old-regexp string i))
      (setq j 0
	    max-j (length new))
      (while (< j max-j)
	(aset string i (aref new j))
	(setq i (1+ i)
	      j (1+ j))))))


;; This could be more efficient with while loops instead of recursion.  Maybe.

;; Returns a value; does not act by side effect.

;; Dies a horrible death if passed a very long string, which is why we use
;; string-replace-regexp-2 instead.
(defun string-substitute-substring-general-case-1 (new old-regexp string)
  (if (string-match old-regexp string)
      (concat (substring string 0 (match-beginning 0))
	      new
	      (string-substitute-substring-general-case
	       new old-regexp (substring string (match-end 0))))
    string))

(defun string-substitute-substring-general-case (new old-regexp string)
  "Call `string-replace-regexp-2'.  Beware special meaning of \\!."
  (string-replace-regexp-2 string old-regexp new))

;; If much replacement is going to happen, this is more efficient.
;; Original version from gaynor@brushfire.rutgers.edu (Silver).
(defun string-replace-regexp-2 (string regexp replacement)
  "Return the string resulting by replacing all of STRING's instances of REGEXP
with REPLACEMENT."
  (save-excursion
    (set-buffer (get-buffer-create " *Temporary*"))
    (erase-buffer)
    (buffer-disable-undo (current-buffer))
    (save-excursion (insert string))
    (while (re-search-forward regexp nil t)
      (replace-match replacement))
    (buffer-string)
    ))

(defun string-replace-regexps (string regexps replacements)
  (save-excursion
    (set-buffer (get-buffer-create " *Temporary*"))
    (erase-buffer)
    (save-excursion (insert string))
    (let (regexp
	  replacement)
      (while regexps
	(setq regexp (car regexps)
	      replacement (car replacements))
	(goto-char (point-min))
	(while (re-search-forward regexp nil t)
	  (replace-match replacement))
	(setq regexps (cdr regexps)
	      replacements (cdr replacements))))
    (buffer-string)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Conversion
;;;

(defun string->integer-default (string &optional default)
  "If STRING represents an integer, return it; otherwise return DEFAULT."
  (let ((result (condition-case nil
		    (car (read-from-string string))
		  (error nil))))
    (if (integerp result)
	result
      default)))
(fset 'string->number-default (symbol-function 'string->integer-default))

(defun string->integer-or-nil (string)
  (string->integer-default string nil))
(fset 'string->number-or-nil (symbol-function 'string->integer-or-nil))

(defun string->integer (string)
  "Return the integer represented by STRING, or err.
See also `string->integer-default'."
  (or (string->integer-or-nil string)
      (error "`%s' doesn't look like an integer." string)))
(fset 'string->number (symbol-function 'string->integer))

;; Backward compatibility
(fset 'string->integer-no-check (symbol-function 'string->integer))
(fset 'string->number-no-check (symbol-function 'string->integer-no-check))

;; The conversions in the other direction are already written.
(fset 'integer->string (symbol-function 'int-to-string))
(fset 'number->string (symbol-function 'int-to-string))

;; Integer or nil
(defun integer-or-nil->string (integer)
  (if (integerp integer)
      (int-to-string integer)
    ""))
(fset 'number-or-nil->string (symbol-function 'integer-or-nil->string))

(defun string-or-nil->integer-or-nil (string-or-nil)
  (and string-or-nil
       (string->integer-or-nil string-or-nil)))
(fset 'string-or-nil->number-or-nil
      (symbol-function 'string-or-nil->integer-or-nil))

(defun string->boolean (string)
  (not (string= string "nil")))

(fset 'i->s 'int-to-string)

;; ;; This is unacceptable because of version 18 bug:  can't have a marker at 0.
;; (defun string-is-integer-p-save-match (string)
;;   "Return t if string looks like the representation of an integer, nil otherwise."
;;   (let ((md (match-data)))
;;     (prog1
;; 	(string-match "^[ 	]*-?[0-9]+" string)
;;       (store-match-data md))))
;; (fset 'string-is-number-p-save-match
;;       (symbol-function 'string-is-integer-p-save-match))
;;
;; (defun string->integer-save-match (integer-string)
;;   (if (string-is-integer-p-save-match integer-string)
;;       (car (read-from-string integer-string))
;;     (error "`%s' doesn't look like an integer." integer-string)))
;; (defun string->number-save-match (symbol-function 'string->integer-save-match))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Atoms
;;;

;; This is a better name.
(fset 'atomp (symbol-function 'atom))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Lists
;;;

(defun flatten (list)
  "Return a list of all of the elements of LIST's elements.
For instance, ((a b) (c) (d (e f) g))) => (a b c d (e f) g)."
  (apply (function append) list))

(defun string-member (item list)
  "Return first link whose car is `string-equal' to ITEM from LIST."
  (let ((ptr list)
        (done nil)
        (result '()))
    (while (not (or done (null ptr)))
      (cond ((string-equal item (car ptr))
             (setq done t)
             (setq result ptr)))
      (setq ptr (cdr ptr)))
    result))

(defun last-element (list)
  "Return last element of LIST."
  (let ((last nil))
    (while list
      (if (null (cdr list))
	  (setq last (car list)))
      (setq list (cdr list)))
    last
    ))

;; Which should come first, N or LIST?
(defun firstn (n list)
  "Return a copy of the first N elements of LIST."
  (let ((result '()))
    (while (and list (> n 0))
      (setq result (cons (car list) result)
	    n (1- n)
	    list (cdr list)))
    (nreverse result)))

;; ;; Don't use this; use mde-add-to-hook instead.
;; (defmacro add-hook-to-list-hack (hook-list hook-func)
;;   "Modify HOOK-LIST to include the HOOK-FUNC only if it is not already
;; a member of the hook list.  The modified hook list is returned,
;; otherwise nil."
;;   (` (if (not (memq (, hook-func) (, hook-list)))
;; 	 (setq (, hook-list) (cons (, hook-func) (, hook-list))))))
;; (make-obsolete 'add-hook-to-list-hack 'mde-add-to-hook)

;;; Emacs provides rassq, but rassoc is nice to have too.
(defun rassoc (elt list)
  "Return non-nil if ELT is the cdr of an element of LIST.  Comparison done with  `equal'.
The value is actually the element of LIST whose cdr is ELT."
  (let (result)
    (while list
      (if (equal elt (cdr (car list)))
	  (setq result (car list)
		list nil)
	(setq list (cdr list))))
    result))


;;; Find patterns in lists.

;; Could have an optional argument which returns non-nil only if partial = 0.
(defun find-pattern (list)
  "Check for a repeated pattern in LIST.
Return a list of (length repeatcount partial) if one is found; nil otherwise.
The pattern is length elements long and occurs fully repeatcount times;
after that, its first partial elements appear.  The length of LIST is
\(+ (* length repeatcount) partial).  Tests are done with `equal'."
  (let* ((len (length list))
	 (pattern-len 1)
	 (max-pattern-len (/ len 2))
	 result)
    (while (<= pattern-len max-pattern-len)
      (if (repeated-pattern-p pattern-len list)
	  (setq result (list pattern-len (/ len pattern-len) (% len pattern-len))
		pattern-len len)
	(setq pattern-len (+ pattern-len 1))))
    result))

;; It is *not* necessary to check the last (- (length list) pattern-length)
;; elements against the first ones; because of transitivity, they will
;; match if this returns t.
(defun repeated-pattern-p (pattern-length list)
  (let ((leader (nthcdr pattern-length list))
	(result t))
    (while leader
      (if (not (equal (car leader) (car list)))
	  (setq result nil
		leader nil)
	(setq leader (cdr leader)
	      list (cdr list))))
    result))

(defun random-nth (list)
  "Return a random element of LIST."
  (nth (% (abs (random)) (length list)) list))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Functions
;;;

;; This can't be a macro because epoch funcalls  (function functionp).
(defun functionp (fun)
  "Return t if FUN is a function, nil otherwise."
  (or (subrp fun)
      (and (symbolp fun)
	   (fboundp fun))
      (and (listp fun)
	   (eq (car fun) 'lambda)
	   (listp (cadr fun)))))
(proclaim-inline functionp)

(defun funcall-maybe (fun &rest args)
  "If FUN is non-nil, apply it to ARGS.  Otherwise return second argument,
which would have been the first argument to which FUN was applied.
FUN should be a funcallable object or nil.  Compare to `maybe-funcall'."
  (if fun
      (apply fun args)
    (if args
	(car args))))
(put 'maybe-funcall 'edebug-form-spec '(function &rest form))

;; Perhaps I should change the order of arguments.
(defun funcall-maybe-default (default fun &rest args)
  "If FUN is non-nil, apply it to ARGS.  Otherwise return DEFAULT.
FUN should be a funcallable object or nil."
  (if fun
      (apply fun args)
    default))
(put 'maybe-funcall 'edebug-form-spec '(form function &rest form))

(defmacro maybe-funcall (fun &rest args)
  "If FUN is non-nil, apply it to ARGS.  Otherwise return nil.
FUN should be a funcallable object or nil.  Compare to `funcall-maybe'."
  (` (funcall-maybe-default nil (, fun) (,@ args))))
(put 'maybe-funcall 'edebug-form-spec '(function &rest form))

;; ;; Instead of this, one should really be using Jamie Zawinski's mapcar.el.
;; (if (not (fboundp 'old-mapcar))
;;     (fset 'old-mapcar (symbol-function 'mapcar)))
;;
;; (defun mapcar (function &rest lists)
;;   "Like `mapcar', but takes an arbitrary number of equal-length lists.
;; The original `mapcar' is available as `old-mapcar'."
;;   (let (result)
;;     (while (car lists)
;;       (setq result (cons (apply func (mapcar (function car) lists)) result)
;;             lists (mapcar (function cdr) lists)))
;;     (nreverse result)))

;; Obviously this could be (easily) generalized to take a list of integers
;; and to try all of those numbers of arguments; but why would I want that?
(defmacro vararg-call (func noargs1 noargs2 &rest args)
  "Apply FUNC to NOARGS1 (an integer), then (if that fails), to NOARGS2
of the ARGS.  -1 means all arguments.  This macro lets you deal with functions
expecting different numbers of arguments in a uniform way.  Since this is a
macro, don't supply something of the form (function foo) as its first argument;
just supply foo itself."
  (let ((noargs (length args))
	nocommon-args
	common-vars
	common-bindings
	thisvar
	(thisargno 0))

    (if (< noargs1 0) (setq noargs1 (- noargs)))
    (if (< noargs2 0) (setq noargs2 (- noargs)))
    (if (not (= (max noargs1 noargs2) noargs))
	(progn
	  (byte-compile-warn "`%s' was vararg-called with a maximum of %d arguments, but you supplied %d."
			   func (max noargs1 noargs2) noargs)
	  (setq args (firstn (max noargs1 noargs2) args))))
    (setq nocommon-args (min noargs1 noargs2))
    (if (= noargs1 noargs2)
	;; aka (` ((, func) (,@ args)))
	(cons func args)
      (while (< thisargno nocommon-args)
	(setq thisargno (1+ thisargno)
	      thisvar (make-symbol (concat "vararg-common-"
					   (int-to-string thisargno)))
	      common-vars (cons thisvar common-vars)
	      common-bindings (cons (list thisvar (car args)) common-bindings)
	      args (cdr args)))
      (setq common-vars (nreverse common-vars)
	    common-bindings (nreverse common-bindings))
      (` (let (, common-bindings)
	   (condition-case err
	       ;; Try calling it with first number of arguments.
	       ((, func) (,@ common-vars)
		(,@ (if (< nocommon-args noargs1) args)))
	     (wrong-number-of-arguments
	      ;; Call it with second number of arguments.
	      ((, func) (,@ common-vars)
	       (,@ (if (< nocommon-args noargs2) args))))
	     (error
	      ;; Otherwise resignal; "while t" makes this work under the
	      ;; debugger (see, eg, the code for the "error" function).
	      (while t
		(signal (car err) (cdr err))))))))))

;; Test cases:
;; (macroexpand '(vararg-call foo 3 1 bar baz bum))
;; (macroexpand '(vararg-call foo 3 5 bar baz bum quux quux2))

;;; May re-evaluate some arguments:  not a good thing.
;; ;; Obviously this could be (easily) generalized to take a list of integers
;; ;; and to try all of those numbers of arguments; but why would I want that?
;; (defmacro vararg-call (function noargs1 noargs2 &rest args)
;;   "Apply FUNCTION to NOARGS1 (an integer), then (if that fails), to NOARGS2
;; of the ARGS.  -1 means all arguments.  This macro lets you deal with functions
;; expecting different numbers of arguments in a uniform way."
;;   (let ((noargs (length args)))
;;     (if (< noargs1 0) (setq noargs1 (- noargs)))
;;     (if (< noargs2 0) (setq noargs2 (- noargs)))
;;     (if (not (= (max noargs1 noargs2) noargs))
;; 	(byte-compile-warn "`%s' was vararg-called with a maximum of %d arguments, but you supplied %d."
;; 			   function (max noargs1 noargs2) noargs))
;;     (` (condition-case err
;; 	;; Try calling it with first number of arguments.
;; 	((, function) (,@ (butlast args (- noargs noargs1))))
;;       (wrong-number-of-arguments
;;        ;; Call it with second number of arguments.
;;        ((, function) (,@ (butlast args (- noargs noargs2)))))
;;       (error
;;        ;; Otherwise resignal; the "while t" makes this work correctly under the
;;        ;; debugger (see, for instance, the code for the "error" function).
;;        (while t
;; 	 (signal (car err) (cdr err))))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Hooks
;;;

;;; This is add-to-hook.el by Joe Wells.  It is included here to avoid the
;;; overhead of yet another file in some Emacs packages.  I've modified the
;;; function name (to avoid conflicts with other packages) and removed the
;;; provide statement.

;; ----------------------------------------------------------------------
;; Created by: Joe Wells, jbw@maverick.uswest.com
;; Created on: Thu Aug 15 15:40:05 1991
;; Last modified by: Joe Wells, jbw@maverick.uswest.com
;; Last modified on: Thu Aug 15 15:43:30 1991
;; Filename: add-to-hook.el
;; Purpose: contains add-to-hook function

(defun mde-add-to-hook (hook-var &rest items)
  "Add to HOOK one or more FUNCTIONS as correctly as possible.
HOOK evaluates to a symbol.

This is designed to work only with hooks that are run by run-hooks.
Because of run-hooks flexibility, the value a hook has can take a variety
of forms.  This function tries to deal with all of those forms.  If you
find a case it does not handle, please tell me.

After this function is done, the hook variable is a list with each element
one function to be run, regardless of what form the value of the hook took
before."
  (or (symbolp hook-var)
      (error "first argument must be symbol"))
  (or (boundp hook-var)
      (set hook-var nil))
  (let ((hook-value (symbol-value hook-var))
	item)
    (if (or (and (symbolp hook-value)
		 ;; should be not eq nil:
		 hook-value)
	    (and (consp hook-value)
		 (eq 'lambda (car hook-value))))
	(setq hook-value (list hook-value)))
    (while (consp items)
      (setq item (car items))
      (or (memq item hook-value)
	  (setq hook-value (cons item hook-value)))
      (setq items (cdr items)))
    (set hook-var hook-value)))

;;(provide 'add-to-hook)
;;----------------------------------------------------------------------


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Files
;;;

(defun filename-sans-extension (filename)
  (if (string-match ".\\.[-a-z]*$" filename)
      (substring filename 0 (1+ (match-beginning 0)))
    filename))

(defun check-file-readable (filename &optional description)
  (if (not (file-readable-p filename))
      (error "Can't read %s `%s'." (or description "file") filename)))
(proclaim-inline check-file-readable)

;; Just a bit of syntactic sugar to sweeten my day.
(defun if-file-exists-p (filename)
  "Return FILENAME if the file exists, nil otherwise."
  (if (file-exists-p filename)
      filename))
(defun if-file-readable-p (filename)
  "Return FILENAME if the file is readable, nil otherwise."
  (if (file-readable-p filename)
      filename))

(defun file-contents (filename)
  "Return the contents of FILENAME as a string."
  (in-buffer (get-buffer-create " file contents")
    (erase-buffer)
    (insert-file-contents filename)
    (prog1
	(buffer-substring (point-min) (point-max))
      (erase-buffer))))

(defun locate-file-with-extensions (filename extensions)
  "Return the name of a readable file starting with FILENAME
or FILENAME's basename and ending with a string in EXTENSIONS, which is a list.
EXTENSIONS may be nil, in which case FILENAME is searched for as is."
  (if extensions
      (let (result)
	(while (and extensions (not result))
	  (setq result (or (if-file-readable-p (concat filename (car extensions)))
			   (if-file-readable-p (concat (filename-sans-extension
							filename)
						       (car extensions))))
		extensions (cdr extensions)))
	result)
    (if-file-readable-p filename)))

(defun locate-file-with-extensions-on-path (filename extensions path)
  "Return the name of a readable file starting with FILENAME
or FILENAME's basename and ending with a string in EXTENSIONS, which is a list.
PATH is a list of strings representing directories to be searched in
order after the current one; they may be relative directories.
Nil means the current directory."
  (or (locate-file-with-extensions filename extensions)
      (let ((filename-directory (file-name-directory filename))
	    (filename-nondirectory (file-name-nondirectory filename))
	    result candidate-directory)
	(while (and path (not result))
	  (setq candidate-directory (if (car path)
					(file-name-as-directory (car path))
				      default-directory)
		path (cdr path)
		result (locate-file-with-extensions
			;; This check is so we return something reasonable,
			;; not because the code requires the simpler form.
			(if (file-name-absolute-p candidate-directory)
			    (concat candidate-directory filename-nondirectory)
			  ;; This probably only works on Unix.
			  (concat filename-directory candidate-directory
				  filename-nondirectory))
			extensions)))
	result)))

(defun locate-file-on-path (filename path)
  "Return the full path of a file named FILENAME located
in the current directory or on PATH, which is a list of directories (strings)
or nil for the current directory."
  (locate-file-with-extensions-on-path filename nil path))


(defun load-file-maybe (file)
  "Load the file FILE of Lisp code, or do nothing if FILE is nil."
  (if file
      (load-file file)))

(defun delete-file-maybe (file)
  "Delete FILE if it exists; do nothing otherwise."
 (if (and file (file-exists-p file))
     (delete-file file)))

(defun same-file-p (file1 file2)
  "Return t if FILE1 and FILE2 are names for the same file."
  (setq file1 (file-resolve-symlink file1)
	file2 (file-resolve-symlink file2))
  (or (equal file1 file2)
      (equal file1 (file-name-nondirectory file2))
      (equal file2 (file-name-nondirectory file1))
      ;; Works for hard links.  If neither file exists, attributes are nil
      ;; and so trivially equal.
      (and (file-exists-p file1) (file-exists-p file2)
	   (equal (file-attributes file1)
		  (file-attributes file2)))))

(defun file-resolve-symlink (file)
  "Return the non-link FILE eventually points to, or FILE if it's not a symbolic link.
This gets in an infinite loop if FILE points into a circular list of symlinks."
  (while (file-symlink-p file)
    (setq file (expand-file-name (car (file-attributes file))
				 (file-name-directory file))))
  file)

;; Joe Wells <jbw@cs.bu.edu> points out some unpleasant interactions of the
;; following procedure with automounters:
;; If you resolve pathnames to contain no symbolic links, i.e. each component
;; of the path names a hard link, and the pathname points into an automounted
;; area, then the automounter may unmount the file system while you are
;; working with it.  It will only remount the file system if you refer to it
;; through the special automounter symbolic link.  The solution is to remove
;; all symbolic links but the special ones that are used by the automounter,
;; but this is non-trivial to do properly.  Also, if somehow Emacs were fed a
;; non-symbolic-link-containing path to an automounted file, the path should
;; be altered to contain the special automounter symbolic link.  This is
;; really hard.

;; Improved version by David Jones <dmjones@theory.lcs.mit.edu>
;; It will probably die if you hit a directory that you cannot read.
(defun journal-resolve-symlink (filename)
  "Return a full pathname for FILE that contains no symbolic links.
This gets in an infinite loop if FILE points into a circular list of symlinks.
Does not handle hard links."
  (and (stringp filename)
       (file-readable-p filename)
       (let ((components (journal-split-path (expand-file-name filename)))
             (result ""))
         (while components
           (setq result (expand-file-name (car components) result))
           (setq components (cdr components))
           (if (file-symlink-p result)
               (setq result (journal-resolve-symlink
                             (expand-file-name (car (file-attributes result))
                                               (file-name-directory result))))))
         result)))

(defun journal-split-path (string)
  (journal-split-string string "/"))

(defun journal-split-string (string pattern &optional regexp)
  (or regexp (setq pattern (regexp-quote pattern)))
  (let (result)
    (while (string-match pattern string)
      (setq result
            (append result (list (substring string 0 (match-beginning 0)))))
      (setq string (substring string (match-end 0))))
    (append result (list string))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Control structure
;;;

(defmacro repeat-until (test &rest body)
  (` (progn
       (,@ body)
       (while (not (, test))
	 (,@ body)))))
(put 'repeat-until 'lisp-indent-hook 1)
(put 'repeat-until 'edebug-form-spec '(&rest form))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Messages
;;;

(defvar use-electric-help-p nil
  "*Non-nil if Emacs programs should use electric help where possible.
Don't set this to a non-nil value unless the ehelp package is available.")

;; Regrettably, this produces big code by including its argument twice.
(defmacro with-electric-help-maybe (&rest body)
  "Similar to `with-electric-help' if  use-electric-help-p  is non-nil;
otherwise like `with-output-to-temp-buffer' with the \"*Help*\" buffer.
Ehelp is loaded if necessary.
BODY is not a thunk (a function of no arguments) but simply a set of forms."
  (` (if use-electric-help-p
	 (progn
	   (require 'ehelp)
	   (with-electric-help
	    (function (lambda ()
			(,@ body)))))
       (with-output-to-temp-buffer "*Help*"
	 (,@ body)))))

;; Originally by Joe Wells <jbw@cs.bu.edu>
(defun best-fit-message (text &optional buffer)
  "Show TEXT in echo area if it fits or in optional BUFFER (default *Message*)."
  (or buffer (setq buffer "*Message*"))
  (save-excursion
    (set-buffer (get-buffer-create " temp printing buffer"))
    (erase-buffer)
    (buffer-disable-undo (current-buffer))
    (insert text)
    (delete-region (point)
		   (progn
		     (skip-chars-backward " \t\n")
		     (point)))
    (cond ((and (< (current-column) (screen-width))
		(progn
		  (beginning-of-line 1)
		  (bobp)))
	   ;; This can't be just buffer, even though that's non-nil,
	   ;; because it might not be an existing buffer.
	   (delete-windows-on (get-buffer-create buffer))
	   (message "%s" (buffer-substring (point-min) (point-max))))
	  (t
	   (with-electric-help-maybe
	    (princ text))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Cursor movement
;;;

;; Written so as to avoid the use of backquote,
;; which invokes a (byte-compiler?) bug.
(defmacro point-after (&rest commands)
  "Return the value of point after executing the COMMANDS.  Doesn't move point."
  (cons 'save-excursion
	(append
	 commands
	 '((point)))))
(put 'point-after 'edebug-form-spec '(&rest form))

(defun forward-line-wrapping (arg)
  "Like forward-line, but wrap around to the beginning of the buffer if
it encounters the end."
  (interactive "p")
  (let ((to-go (forward-line arg)))
    (cond ((or (plusp to-go) (not (bolp)))
	   (goto-char (point-min))
	   (forward-line-wrapping to-go))
	  ((minusp to-go)
	   (goto-char (point-max))
	   (forward-line-wrapping (1+ to-go))))))

(defun current-line ()
  "Return the line number of the line containing point."
  (save-restriction
    (widen)
    (save-excursion
      (beginning-of-line)
      (1+ (count-lines 1 (point))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Buffers
;;;

;; (defmacro in-buffer (buffer &rest body)
;;   "Executes, in BUFFER, forms of BODY."
;;   (` (let ((this-buffer (current-buffer)))
;;        (set-buffer (get-buffer-create (, buffer)))
;;        (unwind-protect
;; 	   (progn (,@ body))
;; 	 ;; non-killed-buffer-p is slight overkill, but someone might have
;; 	 ;; set this-buffer to something besides a buffer.
;; 	 (if (non-killed-buffer-p this-buffer)
;; 	     (set-buffer this-buffer))))))

;;; This macro duplicates BODY.  This should be fixed.
;; This version, which works when body moves point in a buffer displayed in
;; a window other than the selected window, is from Joe Wells
;; <jbw@cs.bu.edu>.  (If Lisp code moves point in a buffer displayed in a
;; window other than the selected window, Emacs kindly restores point in
;; the buffer to its window's version of point.)
(defmacro in-buffer (buffer &rest body)
  "Executes, in BUFFER, forms of BODY."
  ;; Need get-buffer-create because BUFFER might be a string.
  (` (let ((target-buffer (get-buffer-create (, buffer)))
	   (this-buffer (current-buffer)))
       (if (eq target-buffer this-buffer)
	   (progn
	     (,@ body))
	 ;; Can't use save-excursion here because we only want to save the
	 ;; current buffer, not its value for point.
	 (unwind-protect
	     (progn
	       (set-buffer target-buffer)
	       (let* ((target-window (get-buffer-window target-buffer))
		      (track-window-point-p
		       (and (not (eq target-window (selected-window)))
			    (eq (window-point target-window) (point)))))
		 (prog1
		     (progn
		       (,@ body))
		   (if (and track-window-point-p
			    ;; *** Do I need this check?
			    (eq (current-buffer) target-buffer)
			    (eq target-window (get-buffer-window target-buffer))
			    (not (eq target-window (selected-window))))
		       (set-window-point target-window (point))))))
	   (if (non-killed-buffer-p this-buffer)
	       (set-buffer this-buffer)))))))
(put 'in-buffer 'lisp-indent-hook 1)
(put 'in-buffer 'edebug-form-spec '(&rest form))


;; Similar tricks can be done with syntax-table and current-local-map.
;; Adapted from code by Joe Wells.
(defun copy-buffer-local-variables (buffer)
  "Copy the values of all of BUFFER's local variables into the current buffer."
  (let ((blv (in-buffer buffer (buffer-local-variables)))
	pair symbol)
    (while (consp blv)
      (setq pair (car blv))
      (setq symbol (car pair))
      (if (memq symbol '(0 buffer-undo-list)) ; 18.57 brain-damage!!!!!
	  nil
	(progn
	  (if (not (and symbol (symbolp symbol))) (error "impossible"))
	  (make-local-variable symbol)
	  (set symbol (cdr pair))))
      (setq blv (cdr blv)))))

;;; This did more than I needed, and it screwed up point.
;; (defmacro in-buffer (buffer &rest body)
;;   "Executes, in BUFFER, forms of BODY."
;;   (` (save-window-excursion
;;        (set-buffer (, buffer))
;;        (,@ body))))

(defun refresh-mode-line ()
  "Refresh the mode line."
  (set-buffer-modified-p (buffer-modified-p)))

(defun non-killed-buffer-p (object)
  "Non-nil if OBJECT is a non-killed buffer."
  (and (bufferp object)
       (buffer-name object)))

(defun buffer-flush-undo-history (&optional buffer)
  "Make BUFFER keep undo info but have an empty undo history.
Default is current buffer."
  (buffer-disable-undo (or buffer (current-buffer)))
  (buffer-enable-undo buffer))

;; By Joe Wells <jbw@cs.bu.edu>
;; derived from generate-new-buffer
(defun generate-new-buffer-name (name)
  (if (not (get-buffer name))
      name
    (let ((count 1)
	  (template (concat name "<%d>"))
	  tempname)
      (catch 'found
	(while t
	  (setq tempname (format template count))
	  (if (not (get-buffer tempname))
	      (throw 'found tempname))
	  (setq count (1+ count)))))))


(defun longest-line-in-region (beg end)
  "Return the length of the widest line in the region.
Actually return the number of characters in its on-screen representation.
When called non-interactively, requires two arguments, the region's beginning and end."
  (interactive "r")
  (save-excursion
    (goto-char (min beg end))
    (if (< end beg) (setq end beg))
    (end-of-line 1)
    (let ((result (current-column)))
      (while (< (point) end)
	(end-of-line 2)
	(if (> (current-column) result)
	    (setq result (current-column))))
      result)))


;; Kill-buffer-hooks originally by Roland McGrath <roland@ai.mit.edu>
(provide 'kill-buffer-hooks)		; avoid loading his version over this

(defvar kill-buffer-hooks nil "\
*Hooks to run when a buffer is killed.
May be a function or a list of functions.  If it is a list,
each element is called in order.  Functions are called with
no arguments, with the buffer to be killed current.")

;; Rename the kill-buffer subr to mde-old-kill-buffer.
(if (not (fboundp 'mde-old-kill-buffer))
    (fset 'mde-old-kill-buffer (symbol-function 'kill-buffer)))

(defun kill-buffer (buffer)
  "Like original `kill-buffer', but runs hooks in  kill-buffer-hooks  first."
  (interactive "bKill buffer: ")
  (in-buffer buffer
    (run-hooks 'kill-buffer-hooks)
    (setq buffer (get-buffer buffer))
    (if buffer
	(mde-old-kill-buffer buffer))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Windows
;;;

;; I'm not sure whether this works if the last line is wrapped.  Not that I care.
;; Likewise for bob-onscreen-p and wrapped first line (is that possible?).

(defun eob-visible-p ()
  (save-excursion
    (let ((ht (window-height (selected-window))))
      (move-to-window-line (- ht 2))
      (end-of-line)
      (eobp))))

(defun bob-visible-p ()
  (save-excursion
    (move-to-window-line 0)
    (beginning-of-line)
    (bobp)))

;; Return number of screen lines between START and END.
(defun count-screen-lines (start end)
  (save-excursion
    (save-restriction
      (narrow-to-region start end)
      (goto-char (point-min))
      (vertical-motion (- (point-max) (point-min))))))

;; Return number of screen lines between START and END; return a negative
;; number if END precedes START.
(defun count-screen-lines-signed (start end)
  (let ((lines (count-screen-lines start end)))
    (if (< start end)
	lines
      (- lines))))

(defun count-lines-signed (start end)
  (let ((lines (count-lines start end)))
    (if (< start end)
	lines
      (- lines))))

(defun buffer-visible-p (buffer-or-name)
  "Return non-nil (actually the window) if BUFFER-OR-NAME is in a window.
This is just syntactic sugar for `get-buffer-window'."
  (get-buffer-window buffer-or-name))
(proclaim-inline buffer-visible-p)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Keys
;;;

;;; Semi-transparent way to meta-ify a key.  This still isn't quite what we
;;; want.  The problem is that some patches to permit 8-bit character sets
;;; to be displayed change meta sequences to escape sequences, even without
;;; changing meta-flag to nil.

;;; This does't work, because meta-flag is t in the implementations I'm
;;; targetting.  (Besides the syntax being wrong for a macro.)
;; (defmacro meta-prefix-ify (keys)
;;   "Set meta bit of first character of KEYS or prepend  meta-prefix-char,
;; depending on  meta-flag."
;;   (if meta-flag
;;       (let ((first-char (aref keys 0)))
;; 	(if (zerop (logand meta-bit first-char))
;; 	    (progn
;; 	      (aset keys 0 (logior meta-bit first-char))
;; 	      keys)
;; 	  (concat (list meta-prefix-char) keys)))
;;     (concat (list meta-prefix-char) keys)))

(defmacro meta-prefix-ify (keys)
  "Prepend  meta-prefix-char  to KEYS, a string."
  (` (concat (list meta-prefix-char) (, keys))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Searching and matching
;;;

;;; No-error versions

(defmacro re-search-forward-no-err (regexp &optional limit)
  "Expands to (re-search-forward (, REGEXP) (, LIMIT) t)."
  (` (re-search-forward (, regexp) (, limit) t)))

(defmacro re-search-backward-no-err (regexp &optional limit)
  "Expands to (re-search-backward (, REGEXP) (, LIMIT) t)."
  (` (re-search-backward (, regexp) (, limit) t)))

(defmacro search-forward-no-err (string &optional limit)
  "Expands to (search-forward (, STRING) (, LIMIT) t)."
  (` (search-forward (, string) (, limit) t)))

(defmacro search-backward-no-err (string &optional limit)
  "Expands to (search-backward (, STRING) (, LIMIT) t)."
  (` (search-backward (, string) (, limit) t)))

;;; Other functions

(defmacro re-search-forward-maybe (regexp &rest args)
  "Like re-search-forward, but if regexp is nil, then return nil."
  (` (let ((re (, regexp)))
       (if re
	   (funcall 're-search-forward re (,@ args))))))
(put 're-search-forward-maybe 'edebug-form-spec '(&rest form))

(defmacro looking-at-string (string)
  "t if text after point is STRING."
  (` (looking-at (regexp-quote (, string)))))

;; From Robert Potter <potter@silver.lcs.mit.edu>
(defun looking-back-at (PAT)
  "t when text before point matches regular expression PAT."
  (save-excursion
    (save-restriction
      (narrow-to-region (point-min) (point))
      (re-search-backward (concat "\\(" PAT "\\)\\'") (point-min) t))))

;;; Skipping regexps

(defmacro skip-regexp-forward (regexp &optional match-no)
  "If point is at REGEXP, move past it and return point;
otherwise return nil.
Point is left at the end of match MATCH-NO if it is specified."
  (` (if (looking-at (, regexp))
	 (goto-char (match-end (or (, match-no) 0))))))

(defmacro skip-regexp-forward-maybe (regexp &optional match-no)
  "Return nil if REGEXP is nil; otherwise like skip-regexp-forward."
  (` (let ((re (, regexp)))
       (if re
	   (skip-regexp-forward re)))))

(defmacro skip-regexp-backward (regexp &optional match-no)
  "If point is after REGEXP, move past it and return point;
otherwise return nil."
  (` (let ((here (point)))
       (if (re-search-backward (, regexp))
	   (if (= here (match-end 0))
	       t
	     (progn
	       (goto-char here)
	       nil))))))

(defmacro skip-regexp-backward-maybe (regexp &optional match-no)
  "Return nil if REGEXP is nil; otherwise like skip-regexp-backward."
  (` (let ((re (, regexp)))
       (if re
	   (skip-regexp-backward re)))))

;;; Skipping strings

;; Is this more efficient than regexp-quote and skip-regexp-forward
;; (which equals looking-at-string and goto-char)?
(defmacro skip-string-forward (string)
  "If point is at STRING, move past it and return non-nil;
otherwise return nil."
  (` (let ((s (, string)))
       (if (empty-string-p s)
	   ;; maybe just return t here.
	   t
	 (if (search-forward s (+ (point) (length s)) t)
	     (goto-char (match-end 0)))))))

;; ;; Note STRING is evaluated twice; bad.
;; (defmacro skip-string-forward (string)
;;   "If point is at STRING, moves past it and return non-nil; otherwise err."
;;   (` (if (not (skip-string-forward-no-err (, string)))
;; 	 (error "Didn't find %s after point." (, string)))))

(defmacro skip-string-backward (string)
  "If point is after STRING, move back past it and return t;
otherwise return nil."
  (` (let ((s (, string)))
       (if (empty-string-p s)
	   ;; maybe just return t here.
	   t
	 (search-backward s (- (point) (length s)) t)))))

;; ;; Note STRING is evaluated twice; bad.
;; (defmacro skip-string-backward (string)
;;   "If point is after STRING, move back past it and return t; otherwise err."
;;   (` (if (not (skip-string-forward-no-err (, string)))
;; 	 (error "Didn't find %s before point." (, string)))))

(defun match-string (n &optional source)
  "Return the string matched by parentheses number N.  If there is a
SOURCE string, return the substring of that string; else, return
substring of the current buffer."
  (cond
   ((stringp source)
    (substring source (match-beginning n) (match-end n)))
   (t (buffer-substring (match-beginning n) (match-end n)))))

(defun match-string-maybe (n &optional source)
  "Like match-string, but return nil if there was no match for parenthesis N."
  (and (match-beginning n)
       (match-string n source)))

;; Mostly a debugging function.
(defun show-match-data (source)
  (list (match-string-maybe 0 source)
	(match-string-maybe 1 source)
	(match-string-maybe 2 source)
	(match-string-maybe 3 source)
	(match-string-maybe 4 source)
	(match-string-maybe 5 source)
	(match-string-maybe 6 source)
	(match-string-maybe 7 source)
	(match-string-maybe 8 source)
	(match-string-maybe 9 source)))

;; Lifted from how-many (aka count-matches), but leaner and meaner
(defun how-many-regexp (regexp)
  "Return number of matches for REGEXP following point.
REGEXP should not match the empty string."
  (let ((count 0))
    (save-excursion
     (while (re-search-forward regexp nil t)
       (setq count (1+ count))))
    count))

(defun how-many-string (string)
  "Return number of matches for STRING following point."
  (let ((count 0))
    (save-excursion
     (while (search-forward string nil t)
       (setq count (1+ count))))
    count))

(defun how-many-regexp-overlapping (regexp)
  "Return number of matches for REGEXP following point, including overlapping ones.
REGEXP should not match the empty string."
  (let ((count 0))
    (save-excursion
     (while (re-search-forward regexp nil t)
       (goto-char (1+ (match-beginning 0)))
       (setq count (1+ count))))
    count))

;; Instead of using match-beginning, I could compute the length of the
;; string and use (backward-char (1- string-length)).
(defun how-many-string-overlapping (string)
  "Return number of matches for STRING following point, including overlapping ones."
  (let ((count 0))
    (save-excursion
     (while (search-forward string nil t)
       (goto-char (1+ (match-beginning 0)))
       (setq count (1+ count))))
    count))

(defun how-many-substring-overlapping (substring target)
  "Return number of matches for SUBSTRING in TARGET, including overlapping ones."
  (let ((ss-regexp (regexp-quote substring))
	(count 0)
	(start -1))
    (while (setq start (string-match ss-regexp target (1+ start)))
      (setq count (1+ count)))
    count))

(defun find-char (char string &optional count)
  "Look for CHAR in STRING; return first index in STRING whose element is CHAR.
If optional arg COUNT is specified, return the COUNTth occurrance."
  (if (not count)
      (setq count 1))
  (let ((index 0)
	(string-length (length string))
	(result nil))
    (while (and (< index string-length) (not result))
      (if (char-equal char (elt string index))
	  (if (= count 1)
	      (setq result index)
	    (setq count (1- count))))
      (setq index (1+ index)))
    result))

(defun find-char-from-end (char string &optional count)
  "Look for CHAR in STRING; return last index in STRING whose element is CHAR.
If optional arg COUNT is specified, return the COUNTth occurrance from the end."
  (if (not count)
      (setq count 1))
  (let ((index (1- (length string)))
	(string-length )
	(result nil))
    (while (and (> index -1) (not result))
      (if (char-equal char (elt string index))
	  (if (= count 1)
	      (setq result index)
	    (setq count (1- count))))
      (setq index (1- index)))
    result))

(defun replace-regexp-noninteractive (regexp replacement &optional delimited)
  "Like `replace-regexp', but doesn't modify mark or the mark ring."
  (if delimited
      (setq regexp (concat "\\\\<" regexp "\\\\>")))
  (while (re-search-forward regexp nil t)
    (replace-match replacement)))
(make-obsolete 'replace-regexp 'replace-regexp-noninteractive)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Insertion
;;;

;; (defun insert-maybe (&rest args)
;;   "Like insert, but ignores any arguments which are nil."
;;   (apply (function insert) (delq nil args)))

;; The interior delq is to catch some nils at compile time.
(defmacro insert-maybe (&rest args)
   "Like insert, but ignores any arguments which are nil."
   (` (apply (function insert) (delq nil (, (cons 'list (delq nil args)))))))
(put 'insert-maybe 'edebug-form-spec '(&rest form))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Buffer substitution
;;;

(defun buffer-substitute (substitutions backward check)
  "Make replacements in the current buffer according to SUBSTITUTIONS.
  SUBSTITUTIONS is list of pairs of strings; the cdr of each pair will be
substituted for the car, in order, unless optional argument BACKWARD is
non-nil, in which case the car is substituted for the cdr and the
substitutions are done in reverse order.
  If optional third argument CHECK is non-nil, the user is warned if any of
the substituted-in strings already appears in the buffer; such a situation
would make substitution, then unsubstitution, not yield a result identical
to the original buffer, since all instances of the substituted-in string
will be assumed on the reverse substitution to have been the result of
replacing a substituted-for string.
  Return nil if CHECK is nil or there were no ambiguities; otherwise
return a list of replacements creating ambiguity."

  (if backward
      (setq substitutions (mapcar (function (lambda (sub-cons)
				     (cons (cdr sub-cons) (car sub-cons))))
				  (reverse substitutions))))
    ;;; Much too tricky, and modified the argument besides.
    ;; (let ((subs (reverse substitutions))
    ;; 	    temp)
    ;; 	(while subs
    ;; 	  (setq temp (caar subs))
    ;; 	  (setcar (car subs) (cdar subs))
    ;; 	  (setcdr (car subs) temp)
    ;; 	  (setq subs (cdr subs)))
    ;; 	(setq substitutions (nreverse substitutions)))

  ;; (message "buffer-substitute:  substitutions = %s" substitutions)

  ;; Should do all checking before any substitutions are done.
  ;; Bad:
  ;;  * any to-string appears in text, unless it's an earlier from-string.
  ;;  * any to-string appears in previous to-string without intervening
  ;;    from-string.  (but then it's just stupidly inefficient)

  ;; Perhaps be able to override checks of the substitutions pairs.  Such
  ;; checks will be hairy anyway because we may create an ambiguity by
  ;; replacing part of a match such that the other part is still in the
  ;; buffer unchanged.  With one-character stuff this is obviously much
  ;; easier.
  ;; Perhaps do the checks by character...?

  ;; Don't want to do checks as we do the substitutions because that leaves
  ;; us in a bad state:  the work is partially done.  We want to let the
  ;; guy know before we start.

  ;; If, in the case of an ambiguity, we're just going to give up anyway,
  ;; then perhaps it isn't so bad to do the checks after part of the work
  ;; is done (except that the work already done would have been a waste of
  ;; time).  So maybe make the check of the pairs a preliminary one and do
  ;; the real check as we go.  But in some cases such checks won't be
  ;; necessary.

  ;; Perhaps if we want checks on the substitution strings themselves, then
  ;; do that separately beforehand and call this with check = nil.

  ;; And hey, searching for one instance of a string is pretty cheap, after
  ;; all.  And I don't expect to be calling this with a truly enormous list
  ;; of substitutions anyway.

  ;; I think I'm being too paranoid here.  In many cases I'm not even going
  ;; to call this with check = t.

  (let (from-string to-string ambiguity ambiguities)
    (while substitutions
      (setq from-string (caar substitutions)
	    to-string (cdar substitutions))
      ;; (message "Substituting %s for %s." to-string from-string)
      (goto-char (point-min))

      (if (and check (search-forward to-string nil t))
	  (progn
	    (setq ambiguity (car substitutions))
	    (goto-char (point-min))))

      (replace-string from-string to-string)

      ;; Don't complain if we didn't actually do any substitution.
      (if ambiguity
	  (progn
	    (if (not (= (point) (point-min)))
		(setq ambiguities (cons ambiguity ambiguities)))
	    (setq ambiguity nil)))

      (setq substitutions (cdr substitutions)))
    ambiguities))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;

;; Ought to have a way to set the default value, too.
(defmacro deflocalvar (&rest args)
  "Like defvar, but defines a buffer-local variable."
  (` (progn
       (defvar (,@ args))
       (make-variable-buffer-local (quote (, (car args)))))))
(put 'deflocalvar 'edebug-form-spec '(&rest form))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Interactive
;;;

;; Use this like:  (interactive (interactive-number "Record number: "))

;; The "N" specifier appears to perform this functionality:  it gives back
;; the numeric prefix, not the raw prefix as the texinfo documentation
;; states.

;; (defmacro interactive-number (prompt)
;;   (` (list (if current-prefix-arg
;; 	       (prefix-numeric-value current-prefix-arg)
;; 	     (string->number (read-from-minibuffer (, prompt)))))))

;; (defun foo (bar)
;;   (interactive "NUrgah:")
;;   (print bar))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Constant functions
;;;

(defun t-function (&rest args)
  t)

(defun nil-function (&rest args)
  nil)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Current date
;;;

(defun current-date ()
  "Return the current date, as a human-readable string."
  (let ((now (current-time-string)))
    (concat (substring now 4 8)
	    (substring now 8 11)
	    (substring now 20))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; File local variables
;;;

(defun operate-on-local-variables (region-function)
  "Apply REGION-FUNCTION to the local-variables region of the buffer.
Return t if a local-variables region was found; REGION-FUNCTION should act
by side effect."
  (goto-char (point-max))
  (search-backward "\n\^L"
		   (max (- (point-max) 3000) (point-min)) 'move)
  (if (search-forward "Local Variables:" nil t)
      (progn
	(beginning-of-line 1)
	(funcall region-function (point) (point-max))
	t)
    nil))


;; This page feed is to defeat local variables processing.
